// Lord have mercy on me
#[macro_export]
macro_rules! plugin_info {
    ($typ:ty, $name:literal, $version:literal, $license:literal) => {
        
        lazy_static::lazy_static!(
            static ref PLUGIN_INFO: $crate::PluginInfo = $crate::PluginInfo::new($name, $version, $license);
        );

        static mut PLUG: Option<$typ> = None as Option<$typ>;

        #[no_mangle]
        unsafe extern "C" fn plug_get_info() -> *const std::ffi::c_void {
            PLUGIN_INFO.get_raw_ptr()
        }

        #[no_mangle]
        unsafe extern "C" fn plug_init(ctx: *const $crate::ContextRaw) {
            //PLUG = std::alloc::alloc(std::alloc::Layout::new::<$typ>()) as *mut $typ;
            PLUG = Some(Plug::new());
            
            let ctx = $crate::Context::new(ctx);
            if let Some(p) = &mut PLUG {
                p.init(ctx)
            } else {
                unreachable!();
            }
        }

        #[no_mangle]
        unsafe extern "C" fn plug_pre_reload() -> *mut $typ {
            //TODO: Untested 
            if let Some(p) = &mut PLUG {
                p.pre_reload();
                return p as *mut $typ;
            }
            unreachable!();
        }

        #[no_mangle]
        unsafe extern "C" fn plug_post_reload(state: *mut $typ) {
            PLUG = Some((*state).clone());

            if let Some(p) = &mut PLUG {
                p.post_reload();
            } else {
                unreachable!();
            }
            //TODO: Untested
        }

        #[no_mangle]
        unsafe extern "C" fn plug_poll(buf: *mut i8, len: std::ffi::c_uint) {
            let mut buf = $crate::CBuffer::from_raw_parts_mut(buf, len as usize);
            if let Some(p) = &mut PLUG {
                if let Err(_e) = p.poll(&mut buf) {
                    // TODO: Handle error maybe?
                }
            } else {
                unreachable!();
            }
        }

        #[no_mangle]
        unsafe extern "C" fn plug_free() {
            if let Some(p) = &mut PLUG {
                // std::alloc::dealloc(PLUG as *mut u8, std::alloc::Layout::new::<$typ>());
                p.free();
            } else {
                unreachable!();
            }
        }

    };
}
